import os
import re
import numpy as np
import matplotlib.pyplot as plt
from PIL import Image
import subprocess
import io
import pillow_avif
import rawpy

def load_image(image_path):
    # Check if file is RAW based on extension
    raw_extensions = {'.cr2', '.nef', '.arw', '.dng', '.orf', '.pef', '.raf', '.raw', '.rw2'}
    ext = os.path.splitext(image_path)[1].lower()
    
    if ext in raw_extensions:
        # Use rawpy for RAW files
        try:
            raw = rawpy.imread(image_path)
            return Image.fromarray(raw.postprocess())
        except Exception as e:
            print(f"Erreur lors de la lecture du fichier RAW : {e}")
            raise
    else:
        # Use PIL for other formats
        try:
            return Image.open(image_path)
        except Exception as e:
            print(f"Erreur lors de la lecture de l'image : {e}")
            raise

def get_image_size_in_kb(image_path):
    return os.path.getsize(image_path) / 1024

def map_quality_to_distance(quality):
    min_quality, max_quality = 30, 100
    min_distance, max_distance = 15.0, 1.0
    quality = max(min_quality, min(quality, max_quality))
    distance = ((max_quality - quality) / (max_quality - min_quality)) * (min_distance - max_distance) + max_distance
    return distance

def save_jpeg_xl_with_cjxl(image, output_path, quality):
    temp_png = output_path.replace('.jxl', '.png')
    image.convert("RGB").save(temp_png, format='PNG', optimize=True)
    
    distance = map_quality_to_distance(quality)
    cjxl_command = f'cjxl "{temp_png}" "{output_path}" -d {distance} --quiet'
    subprocess.run(cjxl_command, shell=True)
    os.remove(temp_png)

def save_and_get_size(image, output_dir, file_name, format, quality=None):
    output_path = os.path.join(output_dir, file_name)
    
    if os.path.exists(output_path):
        # Le fichier existe déjà, retourner sa taille
        return get_image_size_in_kb(output_path)
    
    if format == 'JPEG':
        image.save(output_path, format='JPEG', quality=quality)
    elif format == 'WEBP':
        image.save(output_path, format='WEBP', quality=quality)
    elif format == 'AVIF':
        image.save(output_path, format='AVIF', quality=quality)
    elif format == 'JPEG XL':
        save_jpeg_xl_with_cjxl(image, output_path, quality)
    
    return get_image_size_in_kb(output_path)

# Fonction principale pour générer les graphiques pour une seule image
def generate_comparative_plots_for_single_image(image_path, output_dir):
    fixed_quality = 100
    fixed_resolution_scale = 1.0
    fixed_format = 'JPEG'
    
    resolution_scales = np.linspace(0.1, 1.0, 10)
    quality_levels = np.linspace(30, 100, 8, dtype=int)
    formats = ['JPEG', 'WEBP', 'AVIF', 'JPEG XL']
    
    try:
        image = load_image(image_path)
    except Exception as e:
        print(f"Erreur lors du traitement de l'image : {e}")
        return None

    resolution_weights = []
    for scale in resolution_scales:
        resized_image = image.resize((int(image.width * scale), int(image.height * scale)))
        ext = 'jpg'
        file_name = f"{os.path.splitext(os.path.basename(image_path))[0]}_res_{int(scale*100)}.{ext}"
        size_kb = save_and_get_size(resized_image, output_dir, file_name, fixed_format, fixed_quality)
        resolution_weights.append(size_kb)

    quality_weights = []
    resized_image = image.resize((int(image.width * fixed_resolution_scale), int(image.height * fixed_resolution_scale)))
    for quality in quality_levels:
        ext = 'jpg'
        file_name = f"{os.path.splitext(os.path.basename(image_path))[0]}_quality_{int(quality)}.{ext}"
        size_kb = save_and_get_size(resized_image, output_dir, file_name, fixed_format, int(quality))
        quality_weights.append(size_kb)

    qualities_to_compare = [30, 50, 70, 80, 100]
    weights = np.zeros((len(qualities_to_compare), len(formats)))
    
    for i, quality in enumerate(qualities_to_compare):
        for j, fmt in enumerate(formats):
            if fmt == 'JPEG':
                ext = 'jpg'
            elif fmt == 'WEBP':
                ext = 'webp'
            elif fmt == 'AVIF':
                ext = 'avif'
            elif fmt == 'JPEG XL':
                ext = 'jxl'
            file_name = f"{os.path.splitext(os.path.basename(image_path))[0]}_{fmt.lower()}_{int(quality)}.{ext}"
            size_kb = save_and_get_size(resized_image, output_dir, file_name, fmt, int(quality))
            weights[i, j] = size_kb
            print(f"Taille de {file_name} : {size_kb:.2f} KB")

    fig, axs = plt.subplots(1, 3, figsize=(18, 5))

    axs[0].plot(resolution_scales * 100, resolution_weights, marker='o')
    axs[0].set_title(f"Résolution vs Poids (Qualité {fixed_quality}%, Format {fixed_format})")
    axs[0].set_xlabel("Résolution (%)")
    axs[0].set_ylabel("Poids (KB)")

    axs[1].plot(quality_levels, quality_weights, marker='o', color='orange')
    axs[1].set_title(f"Qualité vs Poids (Résolution {int(fixed_resolution_scale*100)}%, Format {fixed_format})")
    axs[1].set_xlabel("Qualité (%)")
    axs[1].set_ylabel("Poids (KB)")

    x = np.arange(len(qualities_to_compare))
    width = 0.18
    offsets = (np.arange(len(formats)) - (len(formats) - 1)/2) * width

    for i, fmt in enumerate(formats):
        axs[2].bar(x + offsets[i], weights[:, i], width, label=fmt)

    axs[2].set_xticks(x)
    axs[2].set_xticklabels([f'Qualité {q}%' for q in qualities_to_compare])
    axs[2].set_xlabel("Qualité (%)")
    axs[2].set_ylabel("Poids (KB)")
    axs[2].set_title(f"Comparaison des formats à même niveau de qualité (Résolution {int(fixed_resolution_scale*100)}%)")
    axs[2].legend()
    axs[2].grid(True)

    plt.tight_layout()

    graph_filename = os.path.join(output_dir, f"{os.path.splitext(os.path.basename(image_path))[0]}_graphs.png")
    plt.savefig(graph_filename)
    plt.close()

    return graph_filename

def generate_html_comparison(output_dir, image_name, image_path):
    # Générer l'image JPEG 100% si elle n'existe pas déjà
    jpg100_filename = f"{image_name}_jpeg_100.jpg"
    jpg100_path = os.path.join(output_dir, jpg100_filename)
    
    if not os.path.exists(jpg100_path):
        try:
            image = load_image(image_path)
            image.save(jpg100_path, format='JPEG', quality=100)
        except Exception as e:
            print(f"Erreur lors de la génération de l'image JPEG 100% : {e}")
            return
    
    # Générer les graphiques
    graph_filename = generate_comparative_plots_for_single_image(image_path, output_dir)
    if graph_filename is None:
        print("Les graphiques n'ont pas pu être générés.")
        graph_relative_path = ''
    else:
        graph_relative_path = os.path.basename(graph_filename)
    
    images_dict = {}
    sizes_dict = {}
    pattern = re.compile(rf'{re.escape(image_name)}_(.+)_(\d+)\.(.+)')
    formats_exclus = {'RES', 'QUALITY'}
    formats_autorises = {'JPEG', 'WEBP', 'AVIF', 'JPEG XL'}
    
    for file_name in os.listdir(output_dir):
        if not os.path.isfile(os.path.join(output_dir, file_name)):
            continue

        match = pattern.match(file_name)
        if not match:
            continue

        format_part = match.group(1).upper()
        quality = match.group(2)
        ext = match.group(3).lower()
        if ext not in ['jpg', 'jpeg', 'webp', 'avif', 'jxl']:
            continue

        if format_part in formats_exclus:
            continue

        key = (format_part, quality)
        file_path = os.path.join(output_dir, file_name)
        images_dict[key] = file_name
        sizes_dict[key] = get_image_size_in_kb(file_path)

    formats = set()
    qualities = set()
    for key in images_dict.keys():
        formats.add(key[0])
        qualities.add(key[1])

    formats = sorted(formats)
    qualities = sorted(qualities, key=lambda x: int(x))

    html_content = f"""<!DOCTYPE html>
<html lang="fr">
<head>
    <meta charset="UTF-8">
    <title>Comparaison des images - {image_name}</title>
    <style>
        .image-wrapper {{
            display: inline-block;
            margin: 10px;
            text-align: center;
        }}
        .image-container {{
            width: 250px;
            height: 250px;
            overflow: hidden;
            position: relative;
            border: 1px solid #ccc;
            background-color: #f0f0f0;
        }}
        .image-container img {{
            position: absolute;
            top: 50%;
            left: 50%;
            max-width: none;
            height: auto;
            width: auto;
        }}
        .image-label {{
            margin-top: 5px;
            font-size: 0.7em;
        }}
        .file-size {{
            font-size: 0.6em;
            color: gray;
        }}
        .comparison-group {{
            margin-bottom: 50px;
        }}
        .group-title {{
            font-weight: bold;
            margin: 20px 0 10px 0;
            font-size: 1.2em;
        }}
        body {{
            font-family: Arial, sans-serif;
            padding: 20px;
        }}
        .clearfix::after {{
            content: "";
            clear: both;
            display: table;
        }}
        .graph-container {{
            text-align: center;
            margin-bottom: 50px;
        }}
        .graph-container img {{
            max-width: 100%;
            height: auto;
        }}
        .original-image-container {{
            text-align: center;
            margin-bottom: 50px;
        }}
        #original-image {{
            max-width: 100%;
            height: auto;
        }}
    </style>
    <link rel="stylesheet" href="https://cdnjs.cloudflare.com/ajax/libs/cropperjs/1.5.12/cropper.min.css">
</head>
<body>
<h1>Comparaison des images - {image_name}</h1>
"""

    if graph_relative_path:
        html_content += f"""
<div class="graph-container">
    <h2>Graphiques d'Analyse</h2>
    <img src="{graph_relative_path}" alt="Graphiques">
</div>
"""

    html_content += f"""
<div class="original-image-container">
    <h2>Sélectionnez la zone de recadrage</h2>
    <img id="original-image" src="{jpg100_filename}" alt="Image originale">
</div>
"""

    for format in formats:
        if format in formats_exclus:
            continue

        html_content += f'<div class="comparison-group">\n'
        html_content += f'<div class="group-title">Comparaison des taux de compression pour le format {format}</div>\n'
        html_content += '<div class="clearfix">\n'
        for quality in qualities:
            key = (format, quality)
            if key in images_dict:
                file_name = images_dict[key]
                image_src = file_name
                file_size = sizes_dict.get(key, 0)
                html_content += f'''
    <div class="image-wrapper">
        <a href="{image_src}" target="_blank">
            <div class="image-container">
                <img src="{image_src}" alt="{format} - Qualité {quality}%">
            </div>
        </a>
        <div class="image-label">Qualité {quality}%</div>
        <div class="file-size">Poids : {file_size:.2f} KB</div>
    </div>
    '''
        html_content += '</div>\n</div>\n'

    for quality in qualities:
        html_content += f'<div class="comparison-group">\n'
        html_content += f'<div class="group-title">Comparaison des formats pour la qualité {quality}%</div>\n'
        html_content += '<div class="clearfix">\n'
        for format in formats:
            if format not in formats_autorises:
                continue

            key = (format, quality)
            if key in images_dict:
                file_name = images_dict[key]
                image_src = file_name
                file_size = sizes_dict.get(key, 0)
                html_content += f'''
    <div class="image-wrapper">
        <a href="{image_src}" target="_blank">
            <div class="image-container">
                <img src="{image_src}" alt="{format} - Qualité {quality}%">
            </div>
        </a>
        <div class="image-label">{format}</div>
        <div class="file-size">Poids : {file_size:.2f} KB</div>
    </div>
    '''
        html_content += '</div>\n</div>\n'

    html_content += """
<!-- Inclusion de Cropper.js JavaScript -->
<script src="https://cdnjs.cloudflare.com/ajax/libs/cropperjs/1.5.12/cropper.min.js"></script>
<script>
document.addEventListener('DOMContentLoaded', function() {
    var image = document.getElementById('original-image');
    var cropper = new Cropper(image, {
        viewMode: 1,
        aspectRatio: 1,
        zoomOnWheel: false,
        ready: function() {
            updateCrops();
        },
        crop: function(event) {
            updateCrops();
        }
    });

    function updateCrops() {
        var cropData = cropper.getData(true);
        var cropX = cropData.x;
        var cropY = cropData.y;
        var cropWidth = cropData.width;
        var cropHeight = cropData.height;

        var imageContainers = document.querySelectorAll('.image-container');
        imageContainers.forEach(function(container) {
            var img = container.querySelector('img');
            var scaleFactor = 250 / cropWidth;
            img.style.width = Math.round(image.naturalWidth * scaleFactor) + 'px';
            img.style.height = Math.round(image.naturalHeight * scaleFactor) + 'px';
            img.style.left = Math.round(-cropX * scaleFactor) + 'px';
            img.style.top = Math.round(-cropY * scaleFactor) + 'px';
        });
    }
});
</script>
</body>
</html>
"""

    output_html = os.path.join(output_dir, f'{image_name}_comparison.html')
    with open(output_html, 'w', encoding='utf-8') as f:
        f.write(html_content)

    print(f"Fichier HTML généré : {output_html}")
    print("Ouvrez ce fichier dans votre navigateur pour voir la comparaison.")

# Utilisation du script
image_path = "chemin/vers/votre/fichier"  # Chemin de votre image
output_dir = "output_images_comparison"  # Dossier contenant les images générées
image_name = os.path.splitext(os.path.basename(image_path))[0]  

# Créer le dossier de sortie s'il n'existe pas
if not os.path.exists(output_dir):
    os.makedirs(output_dir)

# Vérifier que le fichier d'image existe
if not os.path.exists(image_path):
    print(f"Le fichier {image_path} n'existe pas.")
else:
    generate_html_comparison(output_dir, image_name, image_path)